from aux.MSIS_security import MSIS_summarize_attacks, MSISParameterSet
from aux.MLWE_security import MLWE_summarize_attacks, MLWEParameterSet
from math import sqrt, log, pi, floor, ceil
from scipy.special import betaincinv, gammaln
from aux.entropy_coordinate_hyperball import compute_entropy_coordinate
from aux.tau_computer import tau_computer
from aux.S_Estimator import S_estimate
import sys

# This script is a modification of the one for Dilithium, available at
# https://github.com/pq-crystals/security-estimates
# Please run python3 HAETAE-estimates.py --help for a list of commands.

#Select which estimators should be ran
#The key recovery estimator is much slower than the others and should be toggled off when k,l and eta are unchanged
weak_uf = True
strong_uf = True
key_recovery = True
size = True
fast_LWE = False

if "--no-weak_uf" in sys.argv:
    weak_uf = False
if "--no-strong_uf" in sys.argv:
    strong_uf = False
if "--no-key_recovery" in sys.argv:
    key_recovery = False
if "--no-size" in sys.argv:
    size = False
if "--fast_LWE" in sys.argv or "-f" in sys.argv:
    fast_LWE = True
if len(sys.argv)<=1 or "--help" in sys.argv:
    print("Security Estimator for HAETAE signature.\nOPTIONS:\n--record_xxx: Adds the record parameter sets for xxx security (xxx can be 120, 180 or 260) to the list of parameters.\n--param=\"n=256 q=122889 k=2 l=2 eta=3 security=120 adapt=1\": adds the following parameter set to the list. Multiple parameter sets can be input this way. security must be 120, 180 or 260. If adapt is set to 1, k, l and eta may be increased until target security is reached.\n--no-weak_uf: skips the weak unforgeability hardness computation.\n--no-strong_uf: skips the strong unforgeability hardness computation.\n--no-key_recovery: skips the key recovery hardness computation.\n--no-size: skips the expected verification key and signature sizes computation.\n--fast_LWE or -f: skips the dual attack for LWE parameters. Makes the script a lot faster. Note that this may overestimate the LWE cost by a few bits.")
    exit()

###########################################################################
######################## PRELIMINARIES ####################################
###########################################################################


#This defines, for each security level, the necessary entropy in the hash function.
#I have yet to understand why it is this way.
entropy = { 120 : 192, 180 : 225, 260 : 257 }

#Loading pre-computed S values and LWE security estimates
#They are stored in a "n k l eta S" manner / "n l q eta bit_security"
#Unpack them and store them in a dictionnary
S_values = dict()
with open('dat/S_values.dat') as f:
    S_values = {(int(line[:-1].split(' ')[0]), int(line[:-1].split(' ')[1]), int(line[:-1].split(' ')[2]), int(line[:-1].split(' ')[3])) : int(line[:-1].split(' ')[4]) for line in f.readlines()}
LWE_values = dict()
with open('dat/LWE_values.dat') as f:
    LWE_values = {(int(line[:-1].split(' ')[0]), int(line[:-1].split(' ')[1]), int(line[:-1].split(' ')[2]), int(line[:-1].split(' ')[3])) : int(line[:-1].split(' ')[4]) for line in f.readlines()}



class HAETAEParameterSet(object):
    def __init__(self, q, n=256, k=1, l=1, S=0, eta=5, tau=0, security=120, which = "gaussian", correct = False):
        """
        This class can be instantiated two ways.
        First, if correct is False, then the input values are used as is. tau and S may still be computed if they are not given as inputs, but this may take some time: it is better to give them when they are known.

        Second, if correct is True, then starting from the input parameters, if they do not give sufficient SIS and LWE security, they will be updated in order to provide enough security.
        Note that it is then up to the user to finish optimizing by changing q. Moreover if the parameters already give sufficient security, the correction will do nothing, even in the case where there is too much security.
        """
        # If tau = 0, it will be computed using the "entropy" dictionary above.
        
        #Type of distribution ("gaussian" or "hyperball")
        self.which = which

        #Ring dimension and modulus
        self.n   = n
        self.q   = q

        #Expected number of iterations (must be consistent with N_lower_bound.py)
        M=4/1.001

        #Entropy of hash function
        self.tau = tau
        if tau == 0:
            self.tau = tau_computer(self.n, entropy[security])
            print("Determining best value for tau: "+str(self.tau)+".\n")

        #LWE parameters
        self.l   = l
        self.eta = eta
        if correct:
            print("Determining best value for (l,eta).")
            self.eta = 1
            print("Trying (l,eta)=("+str(self.l)+","+str(self.eta)+")...")
            if not ((self.n, self.l,self.eta,self.q) in LWE_values.keys()):
                LWE_values[self.n, self.l, self.eta, self.q] = MLWE_summarize_attacks(MLWEParameterSet(self.n, max(self.l-1,1), max(self.l-1,1), self.eta, q, distr="uniform"), fast_LWE)[1]
            LWE_security = LWE_values[self.n, self.l, self.eta, self.q]
            while (LWE_security < security):
                print("Insufficient security, increasing (l,eta).\n")
                self.eta += 1
                if(self.eta>eta):
                    self.eta = 1
                    self.l  += 1
                print("Trying (l,eta)=("+str(self.l)+","+str(self.eta)+")...")
                if not ((self.n, self.l,self.eta,self.q) in LWE_values.keys()):
                    LWE_values[self.n, self.l, self.eta, self.q] = MLWE_summarize_attacks(MLWEParameterSet(self.n, max(self.l-1,1), max(self.l-1,1), self.eta, q, distr="uniform"), fast_LWE)[1]
                else:
                    print("\nBit security for these parameters previously computed.\n")
                LWE_security = LWE_values[self.n, self.l, self.eta, self.q]
            print("Best value for (l,eta) found.\n")

        #SIS parameters
        if correct:
            self.k = self.l-1
        else:
            self.k = k
        self.S = S
        if self.S == 0:
            if (self.n, self.k,self.l,self.eta) in S_values.keys():
                self.S = S_values[(self.n, self.k,self.l,self.eta)]
            else:
                print("Value not found in the list for S. Computing it now.")
                self.S = ceil(S_estimate(100, self.n, self.l+self.k, self.eta))
                S_values[(self.n, self.k,self.l,self.eta)] = self.S
            print("Determining best value for S: "+str(self.S)+".\n")
        self.beta = self.S*sqrt(self.tau)
        #if(which=="gaussian"):
        #    self.sigma = self.beta/(2*sqrt(log(M)))
        #    self.delta = log(self.sigma,2)
        #    #Take into account the growth due to the hint
        #    #############################################
        #    # IMPORTANT -> could be lowered with experiments as in Dilithium-G?
        #    #############################################
        #    self.B = 1.05*sqrt((self.l+self.k)*self.n)*self.sigma+(2)*sqrt(self.k*self.n)*self.sigma/2.
        if(which=="hyperball"):
            self.m = self.n*(self.k+self.l)
            self.gamma1 = self.beta/sqrt((M/2)**(2/self.m)-1)
            self.alpha = floor(self.gamma1*sqrt(self.m)/(4*(self.m+1)))*4
            #Same
            self.B = self.gamma1+sqrt(self.k*self.n)*self.alpha/4+sqrt(self.m)/2


        if correct:
            print("Determining best value for k and updating the other values.")
            current_security = MSIS_summarize_attacks(MSISParameterSet(self.n, self.k + self.l, self.k, self.B, self.q, norm="l2"))[1]
            while (current_security < security and 0!=current_security):
                print("\nIncreasing k to "+str(self.k+1)+".")
                self.k   += 1
                if (self.n, self.k,self.l,self.eta) in S_values.keys():
                    self.S = S_values[(self.n, self.k,self.l,self.eta)]
                else:
                    print("Value not found in the list for S. Computing it now.")
                    self.S = ceil(S_estimate(100, self.n, self.l+self.k, self.eta))
                    S_values[(self.n, self.k,self.l,self.eta)] = self.S
                print("Updating S to "+str(self.S)+".")
                self.beta = self.S*sqrt(self.tau)
                #if(which=="gaussian"):
                #    self.sigma = self.beta/2.3548 #computed with M=4
                #    self.delta = log(self.sigma,2)
                #    self.B     = 1.05*sqrt((self.l+self.k)*self.n)*self.sigma+sqrt(self.k*self.n)*self.sigma/2.
                if(which=="hyperball"):
                    self.m = self.n*(self.k+self.l)
                    self.gamma1 = self.beta/sqrt((M/2)**(2/self.m)-1)
                    self.B = self.gamma1*(1+sqrt(self.k*self.n)/(2*sqrt(self.m)))+sqrt(self.m)/2
                current_security = MSIS_summarize_attacks(MSISParameterSet(self.n, self.k + self.l, self.k, self.B, self.q, norm="l2"))[1]
            if(current_security!=0):
                print("Best value for k found.\n")
            else:
                print("WARNING: could not find correct k. Maybe q is too small?")



        #Parameters for truncation
        #if which=="gaussian":
        #    self.alpha = self.sigma
        elif which=="hyperball":
            self.alpha = floor(self.gamma1*sqrt(self.m)/(4*(self.m+1)))*4
        self.gamma2 = (2*self.q-1)/self.alpha

        # SIS ell_2 bound for unforgeability, using selftargetMSIS
        self.zeta = self.B
        # SIS ell_2 bound for strong unforgeability
        self.zeta_prime = 2*self.B
        if(self.zeta>self.q):
            print("Warning: B > q, which may lead to improved attacks")



# Schemes parameters
####################
# Lowest values for q -> 6657, 5633, 3585, 2561, 1537, 1025.
n= 256

variants             = ["hyperball"]
all_params_gaussian  = []
all_params_hyperball = []

#Parse input parameters
for s in sys.argv:
    if s[:len("--param=")] == "--param=":
        args = s[len("--param="):].split(' ') #Remove the beginning and final "
        print("Currently parsing",args,"\n")
        values = {"q" : "6657", "n" : "256", "k" : "1", "l" : "1", "security" : "120", "which" : "hyperball", "eta" : "2", "adapt" : "0"}
        for opt in args:
            values[opt.split("=")[0]] = opt.split("=")[1]
        if not values["which"] in variants:
            variants += [values["which"]]
        #if values["which"] == "gaussian":
        #    all_params_gaussian  += [("Gaussian Module HAETAE " +str(s), HAETAEParameterSet(int(values["q"]), int(values["n"]), k=int(values["k"]), l=int(values["l"]), eta=int(values["eta"]), security=int(values["security"]), which=values["which"], correct=bool(int(values["adapt"]))))]
        elif values["which"] == "hyperball":
            all_params_hyperball += [("Hyperball Module HAETAE "+str(s), HAETAEParameterSet(int(values["q"]), int(values["n"]), k=int(values["k"]), l=int(values["l"]), eta=int(values["eta"]), security=int(values["security"]), which=values["which"], correct=bool(int(values["adapt"]))))]
        else:
            raise TypeError("Unrecognized signature type: expected hyperball and got: "+values["which"])



#Hardcoded parameters.
#if "--gaussian" in sys.argv:
#    #Those are outdated
#    all_params_gaussian += [("Gaussian Module HAETAE Medium "+str(pri), HAETAEParameterSet(pri, n, k=2, l=2, eta=2, security = 120, which = "gaussian", correct = True)) for pri in [12289, 6657, 3585, 1537, 1025]
#                          ]
#    if not "gaussian" in variants:
#        variants += ["gaussian"]

if "--record_120" in sys.argv:
    #Entre 57857 et 67000
    all_params_hyperball += [("Hyperball Module HAETAE Medium", HAETAEParameterSet(64513, n, k=2, l=4, eta=1, security = 120, which = "hyperball", correct = False))]
if "--record_180" in sys.argv:
    all_params_hyperball += [("Hyperball Module HAETAE Recommended", HAETAEParameterSet(64513, n, k=3, l=6, eta=1, security = 180, which = "hyperball", correct = False))]
if "--record_260" in sys.argv:
    all_params_hyperball += [("Hyperball Module HAETAE Very High", HAETAEParameterSet(64513, n, k=4, l=7, eta=1, security=260, which="hyperball", correct = False))]


all_params = { "gaussian" : all_params_gaussian,
               "hyperball" : all_params_hyperball}


#########################
# Conversion to MSIS/MLWE
#########################

def HAETAE_to_MSIS(dps, strong_uf = False):
    if strong_uf:
        return MSISParameterSet(dps.n, dps.k + dps.l, dps.k, dps.zeta_prime, dps.q, norm="l2")
    return MSISParameterSet(dps.n, dps.k + dps.l, dps.k, dps.zeta, dps.q, norm="l2")


def HAETAE_to_MLWE(dps):
    #Also valid for Ring case as solving the NTRU instance is solving hf+g=0, an LWE instance
    return MLWEParameterSet(dps.n, max(dps.l-1,1), dps.k, dps.eta, dps.q, distr="uniform")

text_SIS = ["BKZ block-size $b$ to break SIS","Best Known Classical bit-cost","Best Known Quantum bit-cost","Best Plausible bit-cost"]
text_LWE = ["BKZ block-size $b$ to break LWE","Best Known Classical bit-cost","Best Known Quantum bit-cost","Best Plausible bit-cost"]


##################
# Size Computation
##################

def HAETAE_Signature_Size(dps):
    """
    Computes the expected size of a signature depending on the type of distribution used.
    """
    #Size of tilde{c} is n bits
    size_c = dps.n 
    if dps.which=="gaussian":
        # See p.17 of Dilithium-G
        size_c_h = size_c + 2.5*dps.n*dps.k
        return size_c_h+(2.25+dps.delta)*dps.n*dps.l
    else:
        m = dps.n*(dps.l)
        #Keep same approximation as Gaussians for shape of h
        size_c_h = size_c + 2.5*dps.n*dps.k 
        return size_c_h + (compute_entropy_coordinate(m,floor(dps.gamma1))+1)*m #Huffman coding of each coordinate.

def HAETAE_Entropy(dps):
    """
    Computes the optimal expected size of a signature by replacing the encoding of z by its entropy
    """
    #Size of tilde{c} is n bits
    size_c = dps.n 
    if dps.which=="gaussian":
        # See p.17 of BLISS-G
        size_c_h = size_c + 2.5*dps.n*dps.k
        return size_c_h+(1.8257+dps.delta)*dps.n*dps.l
    else:
        m = dps.n*(dps.l+dps.k)
        #Keep same approximation as Gaussians for shape of h
        size_c_h = size_c + 2.5*dps.n*dps.k #Number of bits necessary to represent the hint, i.e. ~High(z_2)
        return size_c_h +(compute_entropy_coordinate(m,floor(dps.gamma1)))*(dps.n*dps.l) #Number of bits necessary to represent z_1, assuming entropic coding

def HAETAE_PK_Size(dps):
    """
    Computes the expected size of a verification key. This does not depend on the distribution used.
    """
    return (256 + dps.k*dps.n*(int(log(dps.q,2)+1))) #The public vector is computed mod q


# rest of script is just formatting
#############################################
######################### ANALYSIS AND REPORT 
#############################################


#Store new values of S in dat/S_values.dat
f = open('dat/S_values.dat', 'w')
for key in S_values.keys():
    f.write(str(key[0])+" "+str(key[1])+" "+str(key[2])+" "+str(key[3])+" "+str(S_values[key])+"\n")
f.close()

f = open('dat/LWE_values.dat', 'w')
for key in LWE_values.keys():
    f.write(str(key[0])+" "+str(key[1])+" "+str(key[2])+" "+str(key[3])+" "+str(LWE_values[key])+"\n")
f.close()


table_weak_SIS   = { "gaussian" : [len(all_params_gaussian)*[0] for i in range(4)],
                     "hyperball" : [len(all_params_hyperball)*[0] for i in range(4)]}
table_strong_SIS = { "gaussian" : [len(all_params_gaussian)*[0] for i in range(4)],
                     "hyperball" : [len(all_params_hyperball)*[0] for i in range(4)]}
table_LWE        = { "gaussian" : [len(all_params_gaussian)*[0] for i in range(4)],
                     "hyperball" : [len(all_params_hyperball)*[0] for i in range(4)]}
table_size       = { "gaussian" : [0 for i in range(len(all_params_gaussian))],
                     "hyperball" : [0 for i in range(len(all_params_hyperball))]}
table_entropy    = { "gaussian" : [0 for i in range(len(all_params_gaussian))],
                     "hyperball" : [0 for i in range(len(all_params_hyperball))]}
table_pk         = { "gaussian" : [0 for i in range(len(all_params_gaussian))],
                     "hyperball" : [0 for i in range(len(all_params_hyperball))]}


#For each selected scheme, build the estimate cost of selected attacks
for distribution in variants:
    j = 0
    for (scheme, param) in all_params[distribution]:
        print("\n"+scheme)
        print(param.__dict__)
        print("")
        if weak_uf:
            print("=== WEAK UF")
            v = MSIS_summarize_attacks(HAETAE_to_MSIS(param))
            print(v[1])
            for i in range(4):
                table_weak_SIS[distribution][i][j] = v[i]
        if strong_uf:
            print("=== STRONG UF")
            v = MSIS_summarize_attacks(HAETAE_to_MSIS(param, strong_uf=True))
            for i in range(4):
                table_strong_SIS[distribution][i][j] = v[i]
        if key_recovery:
            print("=== SECRET KEY RECOVERY")
            v = MLWE_summarize_attacks(HAETAE_to_MLWE(param),fast_LWE)
            for i in range(4):
                table_LWE[distribution][i][j] = v[i]
        if size:
            print("=== SIGNATURE SIZE")
            table_size[distribution][j] = HAETAE_Signature_Size(param)
            print(table_size[distribution][j])
            table_entropy[distribution][j] = HAETAE_Entropy(param)
            print(table_entropy[distribution][j])
            print("=== PK SIZE")
            table_pk[distribution][j] = HAETAE_PK_Size(param)
            print(table_pk[distribution][j])
        j+=1


for distribution in variants:
    print(distribution.upper()+" HAETAE TABLE")
    print("========================")
    print("\\hline")
    print("$q$"+"".join([" & "+str(dps[1].q) for dps in all_params[distribution]]))
    print("\\\\")
    print("$B$"+"".join([" & "+str(sqrt(dps[1].gamma1**2+dps[1].beta**2)) for dps in all_params[distribution]]))
    print("\\\\")
    print("$B'$"+"".join([" & "+str(dps[1].gamma1) for dps in all_params[distribution]]))
    print("\\\\")
    print("$B''$"+"".join([" & "+str(dps[1].zeta) for dps in all_params[distribution]]))
    print("\\\\")
    print("$(k,\\ell)$"+"".join([" & ("+str(dps[1].k)+","+str(dps[1].l)+")" for dps in all_params[distribution]]))
    print("\\\\")
    print("$\\eta$"+"".join([" & "+str(dps[1].eta) for dps in all_params[distribution]]))
    print("\\\\")
    print("$\\tau$"+"".join([" & "+str(dps[1].tau) for dps in all_params[distribution]]))
    print("\\\\")
    print("$\\alpha$"+"".join([" & "+str(dps[1].alpha) for dps in all_params[distribution]]))
    print("\\\\")
    print("\\hline")
    for j in range(4):
        print(text_SIS[j]+"".join([" & "+str(table_weak_SIS[distribution][j][i])+" ("+str(table_strong_SIS[distribution][j][i])+")" for i in range(len(all_params[distribution]))]))
        print("\\\\")
    print("\\hline")
    for j in range(4):
        print(text_LWE[j]+"".join([" & "+str(table_LWE[distribution][j][i]) for i in range(len(all_params[distribution]))]))
        print("\\\\")
    print("\\hline")
    #print("Expected signature size"+"".join([" & "+str(int(table_size[distribution][i]/8)) for i in range(len(all_params[distribution]))]))
    #print("\\\\")
    print("Signature size with rANS"+"".join([" & "+str(int(table_entropy[distribution][i]/8)) for i in range(len(all_params[distribution]))]))
    print("\\\\")
    print("Expected public key size"+"".join([" & "+str(int(table_pk[distribution][i]/8)) for i in range(len(all_params[distribution]))]))
    print("\\\\")
    print("Sum"+"".join([" & "+str(int((table_pk[distribution][i]+table_entropy[distribution][i])/8)) for i in range(len(all_params[distribution]))]))
    print("\\\\")
    print("\\hline")
    print("========================")


